package org.naokishibata.armarker;

import java.util.*;
import java.util.concurrent.*;
import java.io.*;
import java.nio.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.imageio.*;
import java.text.*;

import org.naokishibata.generic.*;
import org.naokishibata.math.*;
import org.naokishibata.sleef.*;
import org.naokishibata.ocvimgio.*;
import org.naokishibata.armarker.*;

import static org.jocl.CL.*;
import org.jocl.*;

public class AccuTag {
    static final boolean debug = true;
    static final boolean profiling = false;
    static final boolean multiqueue = true;

    static final int nWorkers = 1;
    static final int NCOORDMAX = 64 * 256;
    long localMemSize;

    public boolean enableSubpixel = true;

    public static class Marker {
	public final Vector3D o, x, y;
	public final long payload;

	Marker(Vector3D o, Vector3D x, Vector3D y, long payload) {
	    this.o = o; this.x = x; this.y = y;
	    this.payload = payload;
	}

	public String toString() {
	    return "[" + payload + ":" + o + ", " + x + ", " + y + "]";
	}
    }

    public interface IplImageFetcher {
	public IplImage fetch();
    }

    public AccuTag(int shift, int iw, int ih, int clDeviceID) throws IOException {
	this(shift, iw, ih, clDeviceID, 0, 0, 0, null);
    }

    public AccuTag(int shift, int iw, int ih, int clDeviceID,
		     double defaultAov, double defaultCamAspect, double defaultCamDistortion, 
		     IplImageFetcher iFetcher) throws IOException {
	SHIFT = shift;
	MAPSIZE = 1 << SHIFT;

	coordBuf = new IntBuffer[nWorkers];

	this.iw = iw;
	this.ih = ih;

	monoImg = new IplImage(iw, ih, 8, 1);

	CL.setExceptionsEnabled(debug);

	long prop = 0;
	if (profiling) prop |= CL_QUEUE_PROFILING_ENABLE;
	cls[0] = new CLHelper(clDeviceID, prop);

	if (multiqueue) {
	    for(int i=1;i<nWorkers;i++) {
		cls[i] = new CLHelper(cls[0], prop);
	    }
	} else {
	    for(int i=1;i<nWorkers;i++) {
		cls[i] = cls[0];
	    }
	}

	cls[0].showInfo();

	//

	kernel_clear7x_0 = makeKernel("clear7_A5.cl", "clear7");

	kernel_pass0 = makeKernel("pass0_C6.cl", "pass0");

	kernel_pass0_mono = makeKernel("pass0m_A0.cl", "pass0");

	kernel_pass1 = makeKernel("pass1_A3.cl", "pass1");
	kernel_pass1m = makeKernel("pass1m_A3.cl", "pass1");

	kernel_pass1a = makeKernel("pass1a_C6.cl", "pass1a");
	kernel_pass1b2 = makeKernel("pass2_D1.cl", "pass1b2");

	kernel_pass3s  = makeKernel("pass3s_D0.cl", "pass3s");

	kernel_pass4_0 = makeKernel("pass4_A5.cl", "pass4");
	kernel_pass4a = makeKernel("pass4a_A5.cl", "pass4a");

	kernel_pass5 = makeKernel("pass5_A5.cl", "pass5");
	kernel_pass6 = makeKernel("pass6_C6.cl", "pass6");

	for(int i=0;i<nWorkers;i++) {
	    clSrcBuf[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_ONLY, iw * ih * 3, null, null);

	    clBuf0[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);
	    clBuf1[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);
	    clBuf2[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);
	    clBuf3[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);
	    clBuf4[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);
	    clBuf5[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);
	    clBuf6[i] = clCreateBuffer(cls[i].context(), CL_MEM_READ_WRITE, MAPSIZE * (ih + 8) * Sizeof.cl_int, null, null);

	    ByteBuffer b = ByteBuffer.allocateDirect(NCOORDMAX * Sizeof.cl_int);
	    b.order(ByteOrder.LITTLE_ENDIAN);
	    coordBuf[i] = b.asIntBuffer();

	    clSetKernelArg(kernel_clear7x_0, 0, Sizeof.cl_mem, Pointer.to(clBuf0[i]));
	    clSetKernelArg(kernel_clear7x_0, 1, Sizeof.cl_mem, Pointer.to(clBuf1[i]));
	    clSetKernelArg(kernel_clear7x_0, 2, Sizeof.cl_mem, Pointer.to(clBuf2[i]));
	    clSetKernelArg(kernel_clear7x_0, 3, Sizeof.cl_mem, Pointer.to(clBuf3[i]));
	    clSetKernelArg(kernel_clear7x_0, 4, Sizeof.cl_mem, Pointer.to(clBuf4[i]));
	    clSetKernelArg(kernel_clear7x_0, 5, Sizeof.cl_mem, Pointer.to(clBuf5[i]));
	    clSetKernelArg(kernel_clear7x_0, 6, Sizeof.cl_mem, Pointer.to(clBuf6[i]));
	    cls[i].run2D(kernel_clear7x_0, MAPSIZE, ih+8, -1, -1); // -> 0, 1, 2, 3, 4, 5, 6

	    cls[i].finish();
	}

	//

	{
	    long[] tmp = new long[1];
	    clGetDeviceInfo(cls[0].device, CL_DEVICE_LOCAL_MEM_SIZE, Sizeof.cl_ulong, Pointer.to(tmp), null);
	    localMemSize = tmp[0];
	}

	//

	iplImageFetcher = iFetcher;

	this.defaultAov = defaultAov;
	this.defaultCamAspect = defaultCamAspect;
	this.defaultCamDistortion = defaultCamDistortion;
    }

    private cl_kernel makeKernel(String fn, String kName) throws IOException {
	String s = null;

	System.out.println("compiling " + fn);

	s = MiscUtils.readWholeFileAsAString(fn);

	s = s.replace("%KERNEL_NAME%", kName);
	s = s.replace("%WIDTH%", "" + iw);
	s = s.replace("%HEIGHT%", "" + ih);
	s = s.replace("%SHIFT%", "" + SHIFT);
	s = s.replace("%EDGE%", "" + EDGE);

	cl_kernel k = cls[0].compile(kName, s, null);

	return k;
    }

    //

    static final int queueSize = nWorkers * 10;
    static final int INIT = 0, RUNNING = 1, EXITTING = 2;

    Worker[] workers = new Worker[nWorkers];

    Object workerMutex = new Object();
    int workerState = INIT, nextWorkerID = 0;
    LinkedBlockingQueue<LinkedList<Marker>> outputQueue = new LinkedBlockingQueue<LinkedList<Marker>>(queueSize);

    public void startWorkers() {
	if (workerState != INIT) return;

	for(int i=0;i<nWorkers;i++) {
	    workers[i] = new Worker(i);
	}

	synchronized(workerMutex) {
	    workerState = RUNNING;
	    workerMutex.notifyAll();
	}

	Runtime.getRuntime().addShutdownHook(new Thread() {
		public void run() {
		    dispose();
		    for(int i=0;i<nWorkers;i++) {
			while(workers[i].th.getState() != Thread.State.TERMINATED) {
			    try {
				Thread.sleep(100);
			    } catch(InterruptedException ex) {}
			}
		    }
		}
	    });
    }

    public void dispose() {
	workerState = EXITTING;

	for(int i=0;i<nWorkers;i++) {
	    workers[i].th.interrupt();
	}

	outputQueue.offer(new LinkedList<Marker>());
    }

    class Worker implements Runnable {
	final int id;
	Thread th;

	Worker(int id) {
	    this.id = id;
	    th = new Thread(this);
	    th.start();
	}

	public void run() {
	    synchronized(workerMutex) {
		while(workerState == INIT) {
		    try {
			workerMutex.wait();
		    } catch(InterruptedException ex) {}
		}
	    }

	    while(workerState == RUNNING) {
		synchronized(workerMutex) {
		    while(nextWorkerID != id && workerState == RUNNING) {
			try {
			    workerMutex.wait();
			} catch(InterruptedException ex) {}
		    }
		}

		if (workerState != RUNNING) break;

		IplImage iimg = iplImageFetcher.fetch();

		synchronized(workerMutex) {
		    nextWorkerID++;
		    if (nextWorkerID == nWorkers) nextWorkerID = 0;
		    workerMutex.notifyAll();
		}

		if (workerState != RUNNING) break;

		LinkedList<Marker> list = recognize(id, iimg, defaultAov, defaultCamAspect, defaultCamDistortion);

		outputQueue.offer(list);
	    }
	}
    }

    public LinkedList<Marker> recognize() {
	LinkedList<Marker> ret = null;

	do {
	    try {
		ret = outputQueue.take();
	    } catch(InterruptedException ex) {}
	} while(workerState == RUNNING && (ret == null || outputQueue.size() > 2));

	return ret;
    }

    //

    long lastTime[] = new long[nWorkers];
    public double[][] partTime = new double[nWorkers][23];
    public double[][] profStartTime = new double[nWorkers][23];
    public double[][] profEndTime = new double[nWorkers][23];
    public int[] nWorkerFrame = new int[nWorkers];

    long baseTimeCL[] = {0, 0, 0, 0}, baseTimeSys[] = {0, 0, 0, 0};

    void checkTime(int id, int n) {
	if (profiling) {
	    if (n == 0) {
		baseTimeSys[id] = System.currentTimeMillis();
	    } else {
		long t = System.currentTimeMillis() - baseTimeSys[id];
		profStartTime[id][n] += t;
	    }
	}
    }

    void checkTime(int id, int n, cl_event ev) {
	if (profiling) {
	    if (n == 0) {
		nWorkerFrame[id]++;
		baseTimeCL[id] = cls[id].startTime(ev);
	    } else {
		long t = cls[id].startTime(ev) - baseTimeCL[id];
		profStartTime[id][n] += t * 1.0e-6;
		t = cls[id].endTime(ev) - baseTimeCL[id];
		profEndTime[id][n] += t * 1.0e-6;
	    }
	}
    }

    public LinkedList<Marker> recognize(IplImage iimg, double aov, double camAspect, double camDistortion) {
	return recognize(0, iimg, aov, camAspect, camDistortion);
    }

    private Object mutexCL = new Object();

    static final int NEVENTMAX = 6;

    private LinkedList<Marker> recognize(int id, IplImage iimg2, double aov, double camAspect, double camDistortion) {
	final int wsx = -1, wsy = -1;
	final int iw = iimg2.width(), ih = iimg2.height();
	final double tanaov = Math.tan(aov / 2);
	IplImage iimg = null;

	checkTime(id, 0);

	if (iimg2.width() != iw || iimg2.height() != ih) {
	    System.err.println("image size does not match");
	    System.exit(-1);
	}

	if (iimg2.colorMode() == 4343634) {
	    CvUtil.cvCvtColor(iimg2, monoImg, CvUtil.CV_RGB2GRAY);
	    iimg = monoImg;
	} else if (iimg2.colorMode() == 1497453127 && iimg2.nChannels() == 1) {
	    iimg = iimg2;
	} else {
	    System.err.println("only RGB or GRAYSCALE image is supported");
	    System.exit(-1);
	}

	int nCoord = MAPSIZE * ih;
	if (nCoord > NCOORDMAX) nCoord = NCOORDMAX;

	LinkedList<Marker> ret = new LinkedList<Marker>();

	ByteBuffer bb = iimg.getDirect();

	cl_event[] ev = new cl_event[NEVENTMAX];

	if (profiling) {
	    for(int i=0;i<NEVENTMAX;i++) {
		ev[i] = new cl_event();
	    }
	}

	clEnqueueWriteBuffer(cls[id].commandQueue(), clSrcBuf[id], CL_FALSE, 0, iw * ih, Pointer.to(bb), 0, null, ev[0]); // -> clSrcBuf

	synchronized(mutexCL) {
	    // Image preprocess
	    // Edge detection / thinning

	    clSetKernelArg(kernel_pass1m, 0, Sizeof.cl_mem, Pointer.to(clSrcBuf[id]));
	    clSetKernelArg(kernel_pass1m, 1, Sizeof.cl_mem, Pointer.to(clBuf1[id]));
	    clSetKernelArg(kernel_pass1m, 2, Sizeof.cl_mem, Pointer.to(clBuf2[id]));
	    clSetKernelArg(kernel_pass1m, 3, Sizeof.cl_mem, Pointer.to(clBuf3[id]));
	    cls[id].run2D(kernel_pass1m, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy, ev[1]); // clSrcBuf -> 1, 2, 3

	    clSetKernelArg(kernel_pass1a, 0, Sizeof.cl_mem, Pointer.to(clBuf1[id]));
	    clSetKernelArg(kernel_pass1a, 1, Sizeof.cl_mem, Pointer.to(clBuf2[id]));
	    clSetKernelArg(kernel_pass1a, 2, Sizeof.cl_mem, Pointer.to(clBuf3[id]));
	    cls[id].run2D(kernel_pass1a, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy); // 1, 2 -> 1, 2, 3

	    clSetKernelArg(kernel_pass1b2, 0, Sizeof.cl_mem, Pointer.to(clBuf3[id]));
	    clSetKernelArg(kernel_pass1b2, 1, Sizeof.cl_mem, Pointer.to(clBuf0[id]));
	    cls[id].run2D(kernel_pass1b2, 0, 0, iw - EDGE*1, ih - EDGE*1, wsx, wsy); // 3 -> 0

	    // CCL

	    clSetKernelArg(kernel_pass3s,  0, Sizeof.cl_mem, Pointer.to(clBuf0[id]));

	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1, ev[2]); // 0 -> 0
	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1);
	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1);
	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1);
	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1);
	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1);
	    cls[id].run2D(kernel_pass3s , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, -1, -1);

	    // Discard small blobs

	    clSetKernelArg(kernel_pass4_0, 0, Sizeof.cl_mem, Pointer.to(clBuf0[id]));
	    clSetKernelArg(kernel_pass4_0, 1, Sizeof.cl_mem, Pointer.to(clBuf4[id]));
	    cls[id].run2D(kernel_pass4_0  , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy, ev[3]); // 0 -> 4

	    clSetKernelArg(kernel_pass4a, 0, Sizeof.cl_mem, Pointer.to(clBuf0[id]));
	    clSetKernelArg(kernel_pass4a, 1, Sizeof.cl_mem, Pointer.to(clBuf4[id]));
	    clSetKernelArg(kernel_pass4a, 2, Sizeof.cl_mem, Pointer.to(clBuf5[id]));
	    cls[id].run2D(kernel_pass4a , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy); // 0, 4 -> 5

	    // Obtain 8 points for each blob

	    clSetKernelArg(kernel_pass5, 0, Sizeof.cl_mem, Pointer.to(clBuf0[id]));
	    clSetKernelArg(kernel_pass5, 1, Sizeof.cl_mem, Pointer.to(clBuf5[id]));
	    clSetKernelArg(kernel_pass5, 2, Sizeof.cl_mem, Pointer.to(clBuf6[id]));
	    cls[id].run2D(kernel_pass5, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy, ev[4]);   // 0, 5 -> 6

	    clSetKernelArg(kernel_pass6, 0, Sizeof.cl_mem, Pointer.to(clBuf0[id]));
	    clSetKernelArg(kernel_pass6, 1, Sizeof.cl_mem, Pointer.to(clBuf5[id]));
	    clSetKernelArg(kernel_pass6, 2, Sizeof.cl_mem, Pointer.to(clBuf6[id]));
	    cls[id].run2D(kernel_pass6, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);   // 0, 5, 6 -> 6
	}

	clEnqueueReadBuffer(cls[id].commandQueue(), clBuf6[id], CL_TRUE, 0, nCoord * Sizeof.cl_int, Pointer.to(coordBuf[id]), 0, null, ev[5]);  // 8 ->

	checkTime(id, 9);

	if (profiling) {
	    cls[id].finish();

	    for(int i=0;i<NEVENTMAX;i++) {
		clWaitForEvents(NEVENTMAX, ev);
	    }

	    for(int i=0;i<NEVENTMAX;i++) {
		checkTime(id, i, ev[i]);
		cls[id].release(ev[i]);
	    }
	}

	if (profiling && nWorkerFrame[id] == 30) {
	    for(int i=0;i<5;i++) {
		System.out.println("" + i + " : " + (profStartTime[id][i+1] - profStartTime[id][i]));
		profStartTime[id][i] = 0;
		profEndTime[id][i] = 0;
	    }
	    System.out.println("5 : " + (profEndTime[id][5] - profStartTime[id][5]));
	    System.out.println("total : " + profEndTime[id][5]);
	    profStartTime[id][5] = 0;
	    profEndTime[id][5] = 0;
	    nWorkerFrame[id] = 0;
	}

	synchronized(mutexCL) {
	    // Clear buffers

	    clSetKernelArg(kernel_clear7x_0, 0, Sizeof.cl_mem, Pointer.to(clBuf0[id]));
	    clSetKernelArg(kernel_clear7x_0, 1, Sizeof.cl_mem, Pointer.to(clBuf1[id]));
	    clSetKernelArg(kernel_clear7x_0, 2, Sizeof.cl_mem, Pointer.to(clBuf2[id]));
	    clSetKernelArg(kernel_clear7x_0, 3, Sizeof.cl_mem, Pointer.to(clBuf3[id]));
	    clSetKernelArg(kernel_clear7x_0, 4, Sizeof.cl_mem, Pointer.to(clBuf4[id]));
	    clSetKernelArg(kernel_clear7x_0, 5, Sizeof.cl_mem, Pointer.to(clBuf5[id]));
	    clSetKernelArg(kernel_clear7x_0, 6, Sizeof.cl_mem, Pointer.to(clBuf6[id]));
	    cls[id].run2D(kernel_clear7x_0, MAPSIZE, ih+8, -1, -1); // -> 0, 1, 2, 3, 4, 5, 6

	    cls[id].flush();
	}

	checkTime(id, 10);

	//

	LinkedHashSet<Node> allNodes = new LinkedHashSet<Node>();

	double ox=-1, oy=-1, xx=0, xy=0, yx=0, yy=0;
	int nInput = 0, nOutput = 0;

	for(int p=0;p<nCoord/16;p++) {
	    if (coordBuf[id].get(p*16) == 0) break;

	    Vector2D[] vt = new Vector2D[16];
	    int[] vatr = new int[16];

	    boolean valid = true;

	    nInput++;

	    int xmin = MAPSIZE, xmax = -1, ymin = MAPSIZE, ymax = -1;
	    for(int i=0;i<8;i++) {
		int x0 = coordBuf[id].get(p*16+8+i) & ((1 << SHIFT)-1), y0 = (coordBuf[id].get(p*16+8+i) >> SHIFT) & ((1 << SHIFT)-1);
		vt[i] = vt[i+8] = new Vector2D(x0, y0);
		vatr[i] = vatr[i+8] = coordBuf[id].get(p*16+i);
		xmin = min(xmin, x0); xmax = max(xmax, x0);
		ymin = min(ymin, y0); ymax = max(ymax, y0);
	    }

	    if (xmin < 1 || ymin < 1 || xmax >= iw-2 || ymax >= ih-2) continue;

	    int size = max(xmax-xmin, ymax-ymin);

	    Node n = new Node(p, vt, new Vector2D(xmin, ymin), new Vector2D(xmax, ymax));

	    allNodes.add(n);
	}

	for(Node n : allNodes) {
	    int[] nn = {-1, -1, -1, -1};

	    {
		double dmax = -1e+100;
		for(int i=0;i<8;i++) {
		    for(int j=0;j<8;j++) {
			double d = Vector2D.distance(n.vt[i], n.vt[j]);
			if (d > dmax) {
			    dmax = d;
			    nn[0] = i; nn[1] = j;
			}
		    }
		}
	    }

	    Vector2D fcenter = Vector2D.midpoint(n.vt[nn[0]], n.vt[nn[1]]);
		    
	    Vector2D north = new Vector2D(n.vt[nn[0]].x - n.vt[nn[1]].x, n.vt[nn[0]].y - n.vt[nn[1]].y);
	    north = north.normalize();

	    Vector2D east = new Vector2D(n.vt[nn[0]].y - n.vt[nn[1]].y, n.vt[nn[1]].x - n.vt[nn[0]].x);
	    east = east.normalize();

	    {
		double dmax2 = -1e+100, dmax3 = -1e+100;
		for(int i=0;i<8;i++) {
		    Vector2D w;

		    w = n.vt[i].minus(fcenter);
		    double c = east.dot(w);

		    if (c > dmax2) {
			dmax2 = c;
			nn[2] = i;
		    }

		    if (-c > dmax3) {
			dmax3 = -c;
			nn[3] = i;
		    }
		}

		double size2 = Vector2D.distance(n.vt[nn[1]], n.vt[nn[0]]);

		double ratio = Math.min(dmax2, dmax3) / size2;

		if (ratio < 0.07) continue;
	    }

	    int white, black, thre;

	    {
		Vector2D wh = n.vt[nn[0]].plus(n.vt[nn[1]].minus(n.vt[nn[0]]).dot(1.0/20));

		black = pixrgb2(bb, (int)wh.y, (int)wh.x);

		Vector2D bl = n.vt[nn[0]].plus(n.vt[nn[0]].minus(n.vt[nn[1]]).dot(1.0/20));

		white = pixrgb2(bb, (int)bl.y, (int)bl.x);

		if (white - black < 25) continue;

		thre = (white + black) / 2;
	    }

	    {
		org.naokishibata.math.Line2D l01 =  new org.naokishibata.math.Line2D(n.vt[nn[0]], n.vt[nn[1]]);
		org.naokishibata.math.Line2D l23 =  new org.naokishibata.math.Line2D(n.vt[nn[2]], n.vt[nn[3]]);

		fcenter = org.naokishibata.math.Line2D.intersection(l01, l23);

		int cpix = pixrgb2(bb, (int)fcenter.y, (int)fcenter.x);

		if (cpix < thre) continue;
	    }

	    Vector2D[] vertex = new Vector2D[4];

	    {
		double d12 = Vector2D.distance(n.vt[nn[1]], n.vt[nn[2]]);
		double d20 = Vector2D.distance(n.vt[nn[2]], n.vt[nn[0]]);
		double d03 = Vector2D.distance(n.vt[nn[0]], n.vt[nn[3]]);
		double d31 = Vector2D.distance(n.vt[nn[3]], n.vt[nn[1]]);

		if (d12 + d20 > d03 + d31) {
		    vertex[0] = n.vt[nn[2]];
		    vertex[2] = n.vt[nn[3]];
		} else {
		    vertex[0] = n.vt[nn[3]];
		    vertex[2] = n.vt[nn[2]];
		}

		if (n.vt[nn[1]].minus(vertex[0]).cross(vertex[0].minus(n.vt[nn[0]])) > 0) {
		    vertex[1] = n.vt[nn[1]];
		    vertex[3] = n.vt[nn[0]];
		} else {
		    vertex[1] = n.vt[nn[0]];
		    vertex[3] = n.vt[nn[1]];
		}
	    }

	    if (enableSubpixel) {
		refineSubpixel(bb, vertex, 1, black, white);
	    }

	    Vector3D[] c3d;
	    int bestq = -1, bestRot = -1, bestReadBits = -1;
	    double dist = NMInst.dist(iw, tanaov);

	    {
		Vector3D vo = NMInst.pos2vec(vertex[0].x, vertex[0].y, iw, ih, dist, camAspect, camDistortion);
		Vector3D vx = NMInst.pos2vec(vertex[1].x, vertex[1].y, iw, ih, dist, camAspect, camDistortion);
		Vector3D vy = NMInst.pos2vec(vertex[3].x, vertex[3].y, iw, ih, dist, camAspect, camDistortion);

		c3d = nmInst.find(vo, vx, vy);

		double maxScore = 1e+100;

		for(int q=0;q<8;q++) {
		    if (c3d[q*3] == null) break;

		    //Vector3D nv = c3d[q*3+1].minus(c3d[q*3+0]).cross(c3d[q*3+2].minus(c3d[q*3+0]));
		    //if (nv.z > 0) continue;

		    Vector3D v3 = c3d[q*3+0].plus(c3d[q*3+1].minus(c3d[q*3+0])).plus(c3d[q*3+2].minus(c3d[q*3+0]));

		    // Read cells

		    boolean bits[] = new boolean[49];

		    {
			for(int iy = 0; iy < 7; iy++) {
			    for(int ix = 0;ix < 7;ix++) {
				Vector3D v;
				Vector2D v2d;

				v = c3d[q*3+0].
				    plus(c3d[q*3+1].minus(c3d[q*3+0]).dot((iy + 0.5) / 7.0)).
				    plus(c3d[q*3+2].minus(c3d[q*3+0]).dot((ix + 0.5) / 7.0));
				v2d = NMInst.vec2pos(v, iw, ih, dist, camAspect, camDistortion);

				double p = bicubic(bb, v2d.x, v2d.y);

				bits[iy * 7 + ix] = p < thre;
			    }
			}
		    }

		    // Decode payload

		    boolean valid = true;
		    int rot = -1;
		    int readBits = 0;

		    {
			for(int i=0;i<7;i++) {
			    if (!bits[i       ]) { valid = false; break; }
			    if (!bits[i*7     ]) { valid = false; break; }
			    if (!bits[i   + 42]) { valid = false; break; }
			    if (!bits[i*7 +  6]) { valid = false; break; }
			}

			if (!valid) continue;

			if (bits[8] && bits[12] && !bits[40]) {
			    rot = 0;
			} else if (bits[12] && bits[40] && !bits[36]) {
			    rot = 1;
			} else if (bits[40] && bits[36] && !bits[8]) {
			    rot = 2;
			} else if (bits[36] && bits[8] && !bits[12]) {
			    rot = 3;
			} else {
			    continue;
			}

			final int[][] bitmap = {
			    {29, 15, 37, 23,  9, 31, 17, 39, 25, 11, 33, 19, 36, 22, 30, 16, 38, 10, 32, 18, 26},
			    { 9, 11, 15, 17, 19, 23, 25, 29, 31, 33, 37, 39,  8, 10, 16, 18, 22, 26, 30, 32, 38},
			    {19, 33, 11, 25, 39, 17, 31,  9, 23, 37, 15, 29, 12, 26, 18, 32, 10, 38, 16, 30, 22},
			    {39, 37, 33, 31, 29, 25, 23, 19, 17, 15, 11,  9, 40, 38, 32, 30, 26, 22, 18, 16, 10},
			};

			for(int i=0;i<21;i++) {
			    if (bits[bitmap[rot][i]]) readBits |= (1 << i);
			}

			if (((crc9(readBits) << 12) | (readBits & 0xfff)) == readBits) {
			} else {
			    valid = false;
			}
		    }

		    if (valid) {
			Vector2D v2 = NMInst.vec2pos(v3, iw, ih, dist, camAspect, camDistortion);
			double score = 0, s;

			s = Vector2D.distance(v2, vertex[2]);
			//s *= s;
			score += s;

			s = Vector2D.distance(NMInst.vec2pos(c3d[q*3+0], iw, ih, dist, camAspect, camDistortion), vertex[0]);
			//s *= s;
			score += s;

			s = Vector2D.distance(NMInst.vec2pos(c3d[q*3+1], iw, ih, dist, camAspect, camDistortion), vertex[1]);
			//s *= s;
			score += s;

			s = Vector2D.distance(NMInst.vec2pos(c3d[q*3+2], iw, ih, dist, camAspect, camDistortion), vertex[3]);
			//s *= s;
			score += s;

			if (score < maxScore) {
			    maxScore = score;
			    bestq = q;
			    bestRot = rot;
			    bestReadBits = readBits;
			}
		    }
		}
	    }

	    if (bestq != -1) {
		/*
		Vector3D vecx = c3d[bestq*3+1].minus(c3d[bestq*3+0]);
		Vector3D vecy = c3d[bestq*3+2].minus(c3d[bestq*3+0]);
		Vector3D vecz = vecx.cross(vecy).normalized();
		*/

		Vector3D vecx = c3d[bestq*3+1].minus(c3d[bestq*3+0]).normalized();
		Vector3D vecy = c3d[bestq*3+2].minus(c3d[bestq*3+0]).normalized();
		Vector3D vecz = vecx.cross(vecy).normalized();

		vecy = vecz.cross(vecx).normalized();

		Vector3D[] v3 = {
		    c3d[bestq*3+0],
		    c3d[bestq*3+1],
		    c3d[bestq*3+1].plus(c3d[bestq*3+2].minus(c3d[bestq*3+0])),
		    c3d[bestq*3+2],
		};

		ret.add(new Marker(v3[3-bestRot], v3[(4-bestRot)&3], v3[(6-bestRot)&3], bestReadBits & 0xfff));
	    }
	}

	checkTime(id,11);

	return ret;
    }

    static class NMInst {
	Vector3D vo, vx, vy;
	double a, b, c, pmp, pmq, pmr;

	double func(double o) {
	    double x = a*o + pmp * Math.sqrt(Math.max(0, 1 - o*o*(1-a*a)));
	    double y = b*o + pmq * Math.sqrt(Math.max(0, 1 - o*o*(1-b*b)));
	    return pmr * (x * x + y * y - 2 * c * x * y - 2);
	}

	double gradient(double x, double h) { return (func(x + h) - func(x)) / h; }
	double gradient2(double x, double h) { return (func(x + h) - 2*func(x) + func(x - h)) / (h * h); }

	double newton(double lo, double hi, double h, int nloop0, int nloop1) {
	    double flo = func(lo);
	    double fhi = func(hi);

	    for(int i=0;i<nloop0;i++) {
		double nx = (lo + hi) / 2;
		double fnx = func(nx);
		if (fnx * flo > 0) {
		    lo = nx; flo = fnx;
		} else {
		    hi = nx; fhi = fnx;
		}
	    }

	    double x = (lo + hi) / 2;

	    for(int i=0;i<nloop1;i++) {
		double f = func(x);
		double fp = gradient(x, h);
		x = x - f / fp;
		if (x < lo) x = lo;
		if (x > hi) x = hi;
	    }

	    return x;
	}

	double newton2(double lo, double hi, double h, int nloop0, int nloop1) {
	    double flo = gradient(lo, h);
	    double fhi = gradient(hi, h);

	    for(int i=0;i<nloop0;i++) {
		double nx = (lo + hi) / 2;
		double fnx = gradient(nx, h);
		if (fnx * flo > 0) {
		    lo = nx; flo = fnx;
		} else {
		    hi = nx; fhi = fnx;
		}
	    }

	    double x = (lo + hi) / 2;

	    for(int i=0;i<nloop1;i++) {
		double f = gradient(x, h);
		double fp = gradient2(x, h);
		x = x - f / fp;
		if (x < lo) x = lo;
		if (x > hi) x = hi;
	    }

	    return x;
	}

	Vector3D[] find(Vector3D vo, Vector3D vx, Vector3D vy) {
	    Vector3D[] ret = new Vector3D[3*8];
	    int p = 0;

	    this.vo = vo;
	    this.vx = vx;
	    this.vy = vy;

	    a = Math.min(1, vx.dot(vo));
	    b = Math.min(1, vy.dot(vo));
	    c = Math.min(1, vx.dot(vy));

	    final double h = -1e-4;

	    for(int i=0;i<4;i++) {
		pmp = (i & 1) == 0 ? 1 : -1;
		pmq = (i & 2) == 0 ? 1 : -1;
		pmr = 1;

		double sb, st;

		sb = ((i & 3) == 0) ? 0.1 : 1;
		st = Math.min(Math.sqrt(Math.min(1/(1-a*a), 1/(1-b*b))), 10000);
		sb += h * 4;
		st -= h * 4;

		double lo = sb;
		double hi = st;
		double glo = gradient(lo, h);

		if (glo < 0) {
		    pmr = -1;
		    glo = -glo;
		} else {
		    pmr = 1;
		}

		double flo = func(lo);
		double fhi = func(hi);
		double ghi = gradient(hi, h);
	    
		double o0 = lo-1, o1 = lo-1;

		if (ghi >= 0) {
		    if (flo > 0) {
			o0 = lo;
		    } else if (fhi < 0) {
			o0 = hi;
		    } else {
			o0 = newton(lo, hi, h, 5, 3);
		    }
		} else {
		    if (flo*fhi < 0) {
			o0 = newton(lo, hi, h, 5, 3);
		    } else if (flo > 0 && fhi > 0) {
			o0 = flo < fhi ? lo : hi;
		    } else {
			double t = newton2(lo, hi, h, 5, 3);
			double ft = func(t);
			if (ft > 0) {
			    o0 = newton(lo,  t , h, 5, 3);
			    o1 = newton(t ,  hi, h, 5, 3);
			} else {
			    o0 = t;
			}
		    }
		}

		if (o0 >= lo) {
		    double o = o0;
		    double x = a*o + pmp * Math.sqrt(Math.max(0, (a*o)*(a*o) + 1 - o*o));
		    double y = b*o + pmq * Math.sqrt(Math.max(0, (b*o)*(b*o) + 1 - o*o));

		    if (Math.abs(x * x + y * y - 2 * c * x * y - 2) < 0.5) {
			ret[p*3+0] = vo.dot(o);
			ret[p*3+1] = vx.dot(x);
			ret[p*3+2] = vy.dot(y);
			p++;
		    }
		}

		if (o1 >= lo) {
		    double o = o1;
		    double x = a*o + pmp * Math.sqrt(Math.max(0, (a*o)*(a*o) + 1 - o*o));
		    double y = b*o + pmq * Math.sqrt(Math.max(0, (b*o)*(b*o) + 1 - o*o));

		    if (Math.abs(x * x + y * y - 2 * c * x * y - 2) < 0.5) {
			ret[p*3+0] = vo.dot(o);
			ret[p*3+1] = vx.dot(x);
			ret[p*3+2] = vy.dot(y);
			p++;
		    }
		}
	    }

	    return ret;
	}

	public static double dist(double iw, double tanaov) {
	    return iw / 2 / tanaov;
	}

	static Vector3D invLenseWarp(Vector3D w, double distance, double aspect, double distortion, double iw, double ih) {
	    w = new Vector3D(w.x, w.y / aspect, w.z);
	    /*
	    {
		double vx = w.x;
		double vy = w.y;
		double d = vx * vx + vy * vy;
		double m = d / (ih * ih) * distortion;
		vx /= 1 + m;
		vy /= 1 + m;
		w = new Vector3D(vx, vy, w.z);
	    }
	    */
	    return new Vector3D(w.x, w.y, distance);
	}

	public static Vector3D pos2vec(double x, double y, double iw, double ih, double dist, double camAspect, double camDistortion) {
	    //return new Vector3D(x - iw/2, y - ih/2, dist).normalized();
	    Vector3D w = new Vector3D(x - iw/2, y - ih/2, dist);
	    w = invLenseWarp(w, dist, camAspect, camDistortion, iw, ih);
	    return w.normalized();
	}

	static Vector3D lenseWarp(Vector3D v, double distance, double aspect, double distortion, double iw, double ih) {
	    Vector3D w = new Vector3D(v.x * distance / v.z, v.y * distance / v.z, distance);
	    /*
	    {
		double vx = w.x;
		double vy = w.y;
		double d = vx * vx + vy * vy;
		double m = d / (ih * ih) * distortion;
		vx *= 1 + m;
		vy *= 1 + m;
		w = new Vector3D(vx, vy, w.z);
	    }
	    */
	    w = new Vector3D(w.x, w.y * aspect, w.z);
	    return w;
	}

	public static Vector2D vec2pos(Vector3D v, double iw, double ih, double dist, double camAspect, double camDistortion) {
	    //return new Vector2D(dist * v.x / v.z + iw/2, dist * v.y / v.z + ih/2);
	    Vector3D w = lenseWarp(v, dist, camAspect, camDistortion, iw, ih);
	    return new Vector2D(w.x + iw/2, w.y + ih/2);
	}
    }

    //

    static int max(int a, int b) { return a > b ? a : b; }
    static int min(int a, int b) { return a < b ? a : b; }

    static double clamp(double x, double min, double max) {
	if (x < min) return min;
	if (x > max) return max;
	return x;
    }

    //

    static class Node {
	LinkedHashSet<Node> children = new LinkedHashSet<Node>();
	final Vector2D ul, lr;
	final Vector2D[] vt;
	final int ptr;
	final double ar;

	int[] nn;
	Node[] cn;

	Node(int ptr, Vector2D[] vt, Vector2D ul, Vector2D lr) {
	    this.ptr = ptr;
	    this.vt = vt;
	    this.ul = ul;
	    this.lr = lr;
	    this.ar = (lr.x - ul.x) / (lr.y - ul.y + 1e-100);
	}
    }

    //

    int pixrgb(ByteBuffer bb, int iy, int ix) {
	return 0xff & (int)bb.get((iy * iw + ix) + 1);
    }

    int pixrgb2(ByteBuffer bb, int y, int x) {
	if (y < 0 || y >= ih-1 || x < 0 || x >= iw-1) return 0;
	return pixrgb(bb, y, x);
    }

    static double cubicSub(double p0, double p1, double p2, double p3, double x) {
	double u, v, w;
	v = p1 - p2;
	w = p3 - p0;
	u = v * 3.0 + w;
	u = u * x   + (-4.0 * v + (p0 - p1 - w));
	u = u * x   + (p2 - p0);
	u = u * x * 0.5 + p1;
	return u;
    }

    double bicubic(ByteBuffer bb, double vx, double vy) {
	if (vy < 1 || vy >= ih-2 || vx < 1 || vx >= iw-2) return 0;

	int ix = (int)vx;
	double x = vx - ix;
	int iy = (int)vy;
	double y = vy - iy;
	
	return cubicSub(cubicSub(pixrgb(bb, iy-1, ix-1), pixrgb(bb, iy-1, ix), pixrgb(bb, iy-1, ix+1), pixrgb(bb, iy-1, ix+2), x),
			cubicSub(pixrgb(bb, iy+0, ix-1), pixrgb(bb, iy+0, ix), pixrgb(bb, iy+0, ix+1), pixrgb(bb, iy+0, ix+2), x),
			cubicSub(pixrgb(bb, iy+1, ix-1), pixrgb(bb, iy+1, ix), pixrgb(bb, iy+1, ix+1), pixrgb(bb, iy+1, ix+2), x),
			cubicSub(pixrgb(bb, iy+2, ix-1), pixrgb(bb, iy+2, ix), pixrgb(bb, iy+2, ix+1), pixrgb(bb, iy+2, ix+2), x), y);
    }

    //

    Vector2D[] refineSubpixelSub(ByteBuffer bb, Vector2D v0, Vector2D v1, int yNumStep, double xStepSize, int xNumStep, int sign, int thre, int range) {
	Vector2D ov, di, dj;

	Vector2D w = v1.minus(v0);
	w = (new Vector2D(-w.y, w.x)).normalized();
	w = w.dot(sign);

	ov = v0.plus(w.dot(xStepSize * ((1 - xNumStep) / 2.0)));
	di = v1.minus(v0).dot(1.0 / (yNumStep+1.0));
	dj = w.dot(xStepSize);

	double sx2 = 0, sx = 0, syx = 0, sy = 0;

	for(int i=1;i<=yNumStep;i++) {
	    Vector2D v = ov.plus(di.dot(i));

	    //

	    double maxsum = -1e+100;
	    int maxindex = -1;
	    double sum = 0;

	    for(int j=0;j<xNumStep;j++) {
		double p = clamp(bicubic(bb, v.x, v.y) - thre, -range, range);

		sum += p;
		if (sum > maxsum) {
		    maxsum = sum;
		    maxindex = j;
		}

		v = v.plus(dj);
	    }

	    double ax = i / (yNumStep+1.0);
	    double ay = xStepSize * (maxindex + 1 - xNumStep / 2.0);

	    sx  += ax;
	    sx2 += ax * ax;
	    syx += ay * ax;
	    sy  += ay;
	}

	double ca, cb;
	double d = 1.0 / (sx2 * yNumStep - sx * sx);

	ca = d * (yNumStep * syx - sx  * sy);
	cb = d * (     -sx * syx + sx2 * sy);

	Vector2D[] ret = { v0.plus(w.dot(cb)), v1.plus(w.dot(ca + cb)) };

	return ret;
    }

    void refineSubpixel(ByteBuffer bb, Vector2D[] v, int sign, int black, int white) {
	org.naokishibata.math.Line2D[] ln = new org.naokishibata.math.Line2D[4];

	for(int i=0;i < 4;i++) {
	    Vector2D[] a = refineSubpixelSub(bb, v[(i + 1) & 3], v[i], 8, 1.0/8, 32, sign, (int)((white + black) / 2), (int)((white - black) / 2));
	    ln[i] = new org.naokishibata.math.Line2D(new Vector2D(a[0].x, a[0].y), new Vector2D(a[1].x, a[1].y));
	}

	for(int i=0;i < 4;i++) {
	    Vector2D w = org.naokishibata.math.Line2D.intersection(ln[i], ln[(i+1)&3]);
	    if (w != null) v[(i + 1) & 3] = new Vector2D(w.x, w.y);
	}

	for(int i=0;i < 4;i++) {
	    v[i] = new Vector2D(v[i].x + 0.5, v[i].y + 0.5);
	}
    }

    //

    static final int POLY = 0xd5;

    static int crc9(int r) {
	r = 0xfff & (r ^ 0x555);
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	return 0x1ff & r;
    }

    static final int[][] bitmap = {
	{29, 15, 37, 23,  9, 31, 17, 39, 25, 11, 33, 19, 36, 22, 30, 16, 38, 10, 32, 18, 26},
	{ 9, 11, 15, 17, 19, 23, 25, 29, 31, 33, 37, 39,  8, 10, 16, 18, 22, 26, 30, 32, 38},
	{19, 33, 11, 25, 39, 17, 31,  9, 23, 37, 15, 29, 12, 26, 18, 32, 10, 38, 16, 30, 22},
	{39, 37, 33, 31, 29, 25, 23, 19, 17, 15, 11,  9, 40, 38, 32, 30, 26, 22, 18, 16, 10},
    };

    //

    private final int iw, ih;
    private final IplImage monoImg;

    private final int SHIFT;
    private final int MAPSIZE;
    private static final int EDGE = 1;
    private static final int modeThreshold = 120;

    private CLHelper[] cls = new CLHelper[nWorkers];
    private NMInst nmInst = new NMInst();

    private cl_mem[] clSrcBuf = new cl_mem[nWorkers];
    private cl_mem[] clBuf0 = new cl_mem[nWorkers];
    private cl_mem[] clBuf1 = new cl_mem[nWorkers];
    private cl_mem[] clBuf2 = new cl_mem[nWorkers];
    private cl_mem[] clBuf3 = new cl_mem[nWorkers];
    private cl_mem[] clBuf4 = new cl_mem[nWorkers];
    private cl_mem[] clBuf5 = new cl_mem[nWorkers];
    private cl_mem[] clBuf6 = new cl_mem[nWorkers];
    private cl_mem[] clBuf7 = new cl_mem[nWorkers];
    private cl_mem[] clBuf8 = new cl_mem[nWorkers];
    private cl_mem[] clBuf9 = new cl_mem[nWorkers];

    private cl_kernel kernel_clear7x_0;
    private cl_kernel kernel_pass0, kernel_pass0_mono, kernel_pass1, kernel_pass1m, kernel_pass1nt, kernel_pass1a, kernel_pass1b2;
    private cl_kernel kernel_pass3s, kernel_pass3p, kernel_pass4_0, kernel_pass4_1, kernel_pass4a, kernel_pass4be, kernel_pass4d;
    private cl_kernel kernel_pass5, kernel_pass6, kernel_pass7;

    private final IntBuffer[] coordBuf;

    //

    private final IplImageFetcher iplImageFetcher;
    private double defaultAov, defaultCamAspect, defaultCamDistortion;
}
