package org.naokishibata.math;

import org.naokishibata.generic.*;

public class MathUtils {
    public static final double epsilon = 1e-10; // 1e-3; // 1e-10;
    public static final double angleEpsilon = 1e-5; // 1e-2; //1e-5;

    public static final boolean epsilonEquals(double a, double b, double eps) {
	return Math.abs(a-b) <= eps;
    }

    public static final boolean epsilonEquals(double a, double b) {
	return Math.abs(a-b) <= epsilon;
    }

    //

    public static VectorN gradientAprx(ScalarField sf, VectorN v, double h) {
	double[] a = new double[v.m()];
	double[] dva = new double[v.m()];

	for(int i=0;i<v.m();i++) {
	    for(int j=0;j<v.m();j++) {
		dva[j] = 0;
		if (j == i) dva[j] = h;
	    }

	    VectorN d = new VectorN(dva);
	    double fxmh = sf.value(v.minus(d));
	    double fxph = sf.value(v.plus(d));
	    a[i] = (fxph - fxmh) / (2*h);
	}

	return new VectorN(a);
    }

    public static VectorN gradient2Aprx(ScalarField sf, VectorN v, double h) {
	double[] a = new double[v.m()];
	double[] dva = new double[v.m()];

	double fx = sf.value(v);

	for(int i=0;i<v.m();i++) {
	    for(int j=0;j<v.m();j++) {
		dva[j] = 0;
		if (j == i) dva[j] = h;
	    }

	    VectorN d = new VectorN(dva);
	    double fxmh = sf.value(v.minus(d));
	    double fxph = sf.value(v.plus(d));
	    a[i] = (fxmh - 2*fx + fxph) / (h*h);
	}

	return new VectorN(a);
    }
}
