package org.naokishibata.generic;

import java.awt.*;
import java.util.*;

public class FormulaLayout implements LayoutManager {
    public interface LayoutFormula {
	public Rectangle getLayout(int w, int h);
    }

    public interface SizeFormula {
	public Dimension getSize(Container c);
    }

    private int preferredSizeX, preferredSizeY;
    private int minimumSizeX, minimumSizeY;

    public FormulaLayout(int px, int py, int mx, int my) {
	preferredSizeX = px; preferredSizeY = py;
	minimumSizeX = mx; minimumSizeY = my;
    }

    private HashMap<Component, LayoutFormula> map = new HashMap<Component, LayoutFormula>();

    public void addLayoutComponent(String name, Component comp) {
    }

    public void removeLayoutComponent(Component comp) {
    }

    public void setFormula(Component comp, LayoutFormula form) {
	map.put(comp, form);
    }

    private SizeFormula formulaPreferred = null;
    private SizeFormula formulaMinimum = null;

    public void setFormulaForPreferredSize(SizeFormula f) {
	formulaPreferred = f;
    }

    public void setFormulaForMinimumSize(SizeFormula f) {
	formulaMinimum = f;
    }

    public Dimension preferredLayoutSize(Container parent) {
	if (formulaPreferred != null) return formulaPreferred.getSize(parent);

	return new Dimension(preferredSizeX, preferredSizeY);
    }

    public Dimension minimumLayoutSize(Container parent) {
	if (formulaMinimum != null) return formulaMinimum.getSize(parent);

	return new Dimension(minimumSizeX, minimumSizeY);
    }

    public void layoutContainer(Container parent) {
        Insets insets = parent.getInsets();
        int maxWidth = parent.getWidth() - (insets.left + insets.right);
        int maxHeight = parent.getHeight() - (insets.top + insets.bottom);

	int nComps = parent.getComponentCount();

	for(int i=0 ; i < nComps ; i++) {
	    Component c = parent.getComponent(i);
	    LayoutFormula f = map.get(c);
	    if (f == null) continue;
	    Rectangle r = f.getLayout(maxWidth, maxHeight);

	    int w = r.width;
	    int h = r.height;

	    Dimension d;

	    d = c.getMaximumSize();
	    if (d != null && w > d.width) w = d.width;
	    if (d != null && h > d.height) h = d.height;

	    d = c.getMinimumSize();
	    if (d != null && w < d.width) w = d.width;
	    if (d != null && h < d.height) h = d.height;

	    c.setBounds(insets.left + r.x, insets.top + r.y, w, h);
        }
    }
}
