package org.naokishibata.generic;

import java.awt.event.*;
import javax.swing.*;
import java.util.regex.*;

public class RegexTextField extends JTextField implements FocusListener, KeyListener {
    private final String pat1, pat2;
    private boolean selectWhenFocused = false;

    private class RegexVerifier extends InputVerifier {
	public boolean verify(JComponent input) {
	    String s = ((JTextField)input).getText();
	    return Pattern.matches(pat2, s);
	}
    }

    public RegexTextField(String patDuringInput, String patAfterInput, int n) {
	super(n);
	this.pat1 = patDuringInput;
	this.pat2 = patAfterInput;
	setInputVerifier(new RegexVerifier());
	addFocusListener(this);
	addKeyListener(this);
    }

    public void keyTyped(KeyEvent ev) {
	try {
	    char c = ev.getKeyChar();

	    if (Character.getType(c) != Character.CONTROL &&
		Character.getType(c) != Character.UNASSIGNED &&
		!ev.isAltDown()) {
		String s = getDocument().getText(0, getDocument().getLength());

		String t = "";

		if (getSelectedText() == null) {
		    try { t = s.substring(0, getCaretPosition()); } catch(IndexOutOfBoundsException e) {}
		    t = t + c;
		    try { t = t + s.substring(getCaretPosition()); } catch(IndexOutOfBoundsException e) {}
		} else {
		    try { t = s.substring(0, getSelectionStart()); } catch(IndexOutOfBoundsException e) {}
		    t = t + c;
		    try { t = t + s.substring(getSelectionEnd()); } catch(IndexOutOfBoundsException e) {}
		}

		if (!Pattern.matches(pat1, t)) {
		    ev.consume();
		    return;
		}
	    }
	} catch(Exception e) {}
    }

    public boolean isAcceptable() {
	try {
	    String s = getDocument().getText(0, getDocument().getLength());
	    return Pattern.matches(pat2, s);
	} catch(Exception e) {
	}

	return false;
    }

    public void focusGained(FocusEvent e) { if (selectWhenFocused) selectAll(); }

    public void focusLost(FocusEvent e) {}

    public void keyReleased(KeyEvent e) {}
    public void keyPressed(KeyEvent e) {}

    public void setSelectWhenFocused(boolean b) { selectWhenFocused = b; }

    public static RegexTextField doubleTextField(int n) {
	return new RegexTextField("[-+]?[0-9]*\\.?(([0-9]*)|([0-9]+((([eE][-+]?)|([eE][-+]?[0-9]+)))?))",
				  "[-+]?[0-9]*\\.?[0-9]+([eE][-+]?[0-9]+)?", n);
    }

    public static RegexTextField positiveDoubleTextField(int n) {
	return new RegexTextField("[+]?[0-9]*\\.?(([0-9]*)|([0-9]+((([eE][-+]?)|([eE][-+]?[0-9]+)))?))",
				  "[+]?[0-9]*\\.?[0-9]+([eE][-+]?[0-9]+)?", n);
    }

    public static RegexTextField intTextField(int n) {
	return new RegexTextField("[-+]?[0-9]*", "[-+]?[0-9]+", n);
    }

    public static RegexTextField positiveIntTextField(int n) {
	return new RegexTextField("[+]?[0-9]*", "[+]?[0-9]+", n);
    }
}
