package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

public class Line2D implements Serializable {
    private final Vector2D u[];

    public Line2D(Vector2D u0, Vector2D u1) {
	u    = new Vector2D[2];
	u[0] = u0;
	u[1] = u1;
    }

    public String toString() {
	return "(Line "+u[0].toString()+"-"+u[1].toString()+")";
    }

    public Vector2D v(int i) {
	return u[i];
    }

    //

    public static boolean parallel(Line2D l, Line2D m, double eps) {
	Vector2D v0 = Vector2D.normalize(Vector2D.sub(l.u[0],l.u[1]));
	Vector2D v1 = Vector2D.normalize(Vector2D.sub(m.u[0],m.u[1]));

	return Math.abs(Vector2D.crossProduct(v0,v1)) < eps;
    }

    public static boolean parallel(Line2D l, Line2D m) {
	return parallel(l, m, MathUtils.angleEpsilon);
    }

    public static Vector2D intersection(Line2D u, Line2D v, double eps) {
	if (u.u[0].epsilonEquals(v.u[0],eps) || u.u[0].epsilonEquals(v.u[1],eps)) return u.u[0];
	if (u.u[1].epsilonEquals(v.u[0],eps) || u.u[1].epsilonEquals(v.u[1],eps)) return u.u[1];

	double d = (v.u[1].x - v.u[0].x) * (u.u[1].y - u.u[0].y) - (v.u[1].y - v.u[0].y) * (u.u[1].x - u.u[0].x);
	if (d == 0) return null;
	double n = (v.u[0].y - u.u[0].y) * (u.u[1].x - u.u[0].x) - (v.u[0].x - u.u[0].x) * (u.u[1].y - u.u[0].y);
	double q = n / d;

	return new Vector2D(v.u[0].x + q * (v.u[1].x - v.u[0].x), v.u[0].y + q * (v.u[1].y - v.u[0].y));
    }

    public static Vector2D intersection(Line2D u, Line2D v) {
	return intersection(u, v, MathUtils.epsilon);
    }

    /*
    public static boolean onSameSide(Line2D l, Vector2D v, Vector2D w) {
	Vector2D x = intersection(l, new Line2D(v,w));
	if (x == null) return true;
	return LineSegment2D.includes(new LineSegment2D(v,w), x) != 2;
    }
    */

    public static double distance(Line2D l, Vector2D v) {
	double d = Vector2D.crossProduct(Vector2D.sub(l.v(1),l.v(0)), Vector2D.sub(l.v(0),v));
	return  d * d / Vector2D.distance(l.v(1),l.v(0));
    }

    public static boolean includes(Line2D l, Vector2D v) {
	return distance(l,v) < MathUtils.epsilon;
    }

    //

    public boolean isParallelTo(Line2D l, double eps) {
	return parallel(this, l, eps);
    }

    public boolean isParallelTo(Line2D l) {
	return parallel(this, l, MathUtils.angleEpsilon);
    }
}
